# Fruit Warrior

## Cel lekcji

Celem lekcji jest zapoznanie uczniów z podstawowymi pojęciami związanymi z programowaniem obiektowym w języku Python.
Uczniowie pomogą zaimplementować grę Fruit Warrior, w której gracz będzie mógł zdobywać punkty poprzez cięcie owoców.
Dzięki tej lekcji uczestnicy poznają podstawy programowania w języku Python.

## Agenda

0. Przygotowanie środowiska (przed zajęciami)
1. Czym jest język programowania?
2. Visual Studio Code - zapoznanie
3. Podstawy - zmienne, typy danych, operatory
4. Wprowadzenie do Pygame
5. Implementacja metod we Fruit Warrior

## Przygotowanie środowiska

Przed zajęciami uczestnicy powinni mieć zainstalowane na swoich komputerach:

- Visual Studio Code
- Rozszerzenie Python dla Visual Studio Code
- Python 3.x
- Pip (menadżer pakietów Pythona)

Po instalacji udajemy się do terminalu i tworzymy nowe środowisko wirtualne:

```bash
python -m venv venv
```

Następnie aktywujemy środowisko wirtualne:

- Windows:

```bash
venv\Scripts\activate
```

- macOS / Linux:

```bash
source venv/bin/activate
```

Teraz możemy zainstalować bibliotekę Pygame z pliku `requirements.txt`:

```bash
pip install -r requirements.txt
```

Projekt możemy uruchomić poleceniem:

```bash
python GameManager.py
```

Wersja bazowa projektu (basic) uruchomi się, aczkolwiek nie będzie działać poprawnie z powodu braku implementacji metod.

Po przetestowaniu projektu jesteśmy gotowi do prowadzenia zajęć.


## Czym jest język programowania?

Język programowania to zbiór reguł i instrukcji, które pozwalają na stworzenie programu komputerowego. Programy te są zbiorem instrukcji, które komputer wykonuje w określonej kolejności. Języki programowania pozwalają na tworzenie programów, które mogą wykonywać różne zadania, takie jak przetwarzanie danych, wyświetlanie informacji na ekranie, czy sterowanie innymi programami.

Program to spis kroków, które komputer wykonuje, aby osiągnąć określony cel. Programy mogą być bardzo proste, takie jak kalkulator, który dodaje dwie liczby, lub bardzo skomplikowane, takie jak system operacyjny, który zarządza zasobami komputera bądź gra komputerowa.

Przykładem takiego spisu kroków, który możemy napotkać na co dzień, jest przepis na ciasto. Przepis na ciasto to zbiór kroków, które musimy wykonać, aby uzyskać gotowe ciasto. 

1. Zbierz składniki: mąkę, jajka, cukier, masło, itp.
2. Wymieszaj składniki w misce.
3. Włóż ciasto do piekarnika.
4. Piecz ciasto przez 30 minut.
5. Wyjmij ciasto z piekarnika.

Podobnie jak przepis na ciasto, program komputerowy to zbiór kroków, które musimy wykonać, aby uzyskać określony wynik. Oczywiście, najczęściej programy komputerowe są bardziej skomplikowane niż przepis na ciasto, ale idea jest taka sama.

Możemy powiedzieć, że język programowania to taki sam język, jak angielski, niemiecki czy francuski, ale zamiast komunikować się z ludźmi, komunikujemy się z komputerem i mówimy mu, co ma zrobić.

Jednym z najpopularniejszych języków programowania jest Python. Python jest językiem programowania, który jest łatwy do nauki i używania. Jest to język wysokiego poziomu, co oznacza, że jest on bardziej zrozumiały dla ludzi niż dla komputerów. Python jest również językiem interpretowanym, co oznacza, że programy napisane w Pythonie są tłumaczone na kod maszynowy w czasie rzeczywistym, a nie kompilowane przed uruchomieniem. Ma to swoje zalety i wady, o których będziemy mówili więcej podczas semestru, ale dla naszych celów jest to idealne rozwiązanie.

## Visual Studio Code - zapoznanie

Visual Studio Code to edytor tekstu, który jest często używany do pisania programów w języku Python. Jest to darmowy edytor, który jest dostępny na wszystkie platformy, takie jak Windows, macOS, czy Linux. Visual Studio Code ma wiele przydatnych funkcji, takich jak podświetlanie składni, automatyczne uzupełnianie kodu, czy integrację z systemem kontroli wersji Git.

Dużą zaletą Visual Studio Code jest to, że jest on bardzo konfigurowalny i można go dostosować do swoich potrzeb. Możemy zainstalować różne rozszerzenia / pluginy, które dodają nowe funkcje do edytora, takie jak obsługa innych języków programowania, czy integracja z różnymi narzędziami.

Środowiko powinno być już zainstalowane na komputerach uczestników. Na prezentacji pokazujemy uczniom główne części edytora, takie jak:

- eksplorator plików - pozwala na przeglądanie plików w projekcie.

- edytor - miejsce, w którym piszemy kod.

- terminal - pozwala na wykonywanie poleceń w konsoli.

Visual Studio Code powinien mieć też zainstalowane rozszerzenie Python, które pozwala na pisanie i uruchamianie programów w języku Python. (Jeśli nie jest zainstalowane, można je zainstalować z oficjalnego sklepu rozszerzeń Visual Studio Code).

## Podstawy - zmienne, typy danych, operatory

### Typy danych

Python ma wiele różnych typów danych, które możemy używać w naszych programach. Oto kilka podstawowych typów danych:

- `int` - liczba całkowita, np. `5`, `10`, `-3`
- `float` - liczba zmiennoprzecinkowa, np. `3.14`, `2.718`
- `str` - łańcuch znaków, np. `"Hello, World!"`, `"Python is awesome!"`
- `bool` - wartość logiczna, `True` lub `False`

### Zmienne

Zmienne to miejsca w pamięci komputera, w których przechowujemy dane. Zmienne pozwalają nam przechowywać informacje, takie jak liczby, tekst, czy obiekty. W Pythonie, aby zadeklarować zmienną, wystarczy przypisać jej wartość. Na przykład:

```python
x = 5
y = "Hello, World!"
```

Python ma to do siebie, że nie musimy deklarować typu zmiennej, ponieważ Python sam wie, jakiego typu jest zmienna. W powyższym przykładzie, zmienna `x` jest typu `int`, czyli liczba całkowita, a zmienna `y` jest typu `str`, czyli łańcuch znaków.

Poprośmy uczestników o zadeklarowanie kilku zmiennych i wypisanie ich wartości na ekranie. Na przykład ich imię, wiek, ulubiony kolor, itp.

```python
name = "John"
age = 25
color = "blue"
```

Uwaga: 
Czemu nie używamy polskich znaków w nazwach zmiennych? Python pozwala na używanie polskich znaków w nazwach zmiennych, ale nie jest to zalecane, ponieważ może to prowadzić do problemów z kodowaniem. Tak samo najlepiej trzymać się angielskich nazw zmiennych, ponieważ jest to standard w programowaniu - pomoże to też uczestnikom w zyskaniu komfortu z angielskim.

### Operatory

Operatory to symbole, które pozwalają na wykonywanie operacji na zmiennych. Takimi operacjami mogą być dodawanie, odejmowanie, mnożenie, dzielenie, itp. Oto kilka podstawowych operatorów w Pythonie:

- `+` - dodawanie
- `-` - odejmowanie
- `*` - mnożenie
- `/` - dzielenie

Pokażmy uczestnikom kilka przykładów użycia operatorów w Pythonie:

```python
x = 5
y = 3

print(x + y) # 8
print(x - y) # 2
print(x * y) # 15
print(x / y) # 1.6666666666666667
```

Python posiada więcej operatorów, takich jak operator modulo `%`, operator potęgowania `**`, czy operator dzielenia całkowitego `//`, ale na razie skupmy się na tych podstawowych.

Czas nas goni, więc przejdziemy od razu do zabawy z grą Fruit Warrior. Napotkamy jeszcze kilka pojęć, takich jak metody, instrukcje warunkowe czy pętle, ale pobieżnie poznamy je w trakcie lekcji.

## Wprowadzenie do Pygame

Pygame to biblioteka do tworzenia gier w języku Python. Pygame pozwala na tworzenie gier 2D, obsługę dźwięku, grafiki, czy sterowanie wejściem od użytkownika. Biblioeka ta jest często używana do nauki programowania, ponieważ jest łatwa w użyciu i ma wiele przykładów dostępnych w internecie.

Bibioteki w Pythonie to zbiory funkcji i klas, które pozwalają na wykonywanie określonych zadań.
Powiedzmy, że chcemy narysować kwadrat na ekranie. W Pythonie musielibyśmy napisać wiele linii kodu, aby osiągnąć ten cel. Pygame pozwala na narysowanie kwadratu jedną linią kodu.
Inne biblioteki rozwiązują inne problemy, jak rozmawianie z bazą danych, czy przetwarzanie obrazów albo używanie sztucznej inteligencji.
Pygame to biblioteka, która pozwala na tworzenie gier komputerowych w języku Python. Pygame dostarcza funkcje do rysowania obiektów na ekranie, obsługi zdarzeń, czy odtwarzania dźwięku.


W naszym projekcie większość kodu związana z obsługą gry jest już napisana, więc uczestnicy nie muszą się martwić o szczegóły implementacyjne. Naszym zadaniem będzie zaimplementowanie kilku metod, które pozwolą na poprawne działanie gry.

## Implementacja metod we Fruit Warrior

Nasza gra jest imitacją popularnej gry Fruit Ninja, w której gracz musi ciąć owoce, aby zdobywać punkty. W naszej grze gracz steruje mieczem, którym musi ciąć owoce (zielone), które pojawiają się na ekranie. Za każde cięcie owoce gracz zdobywa punkty, a jeżeli przetnie bombę (czerwone), to traci życie.

Jak zaprezentujemy uczestnikom kod naszej gry może okazać się trochę przytłaczający, więc wytłumaczmy pobieżnie, co robią poszczególne pliki.

- GameManager.py - plik główny, który odpowiada za uruchomienie gry i obsługę zdarzeń.
- GameState.py - plik, który przechowuje stan gry.
- Particle.py - plik, który odpowiada za obsługę cząsteczek pochądzących z miecza podczas cięcia.
- Sliceable - klasa, która reprezentuje obiekty, które można przeciąć (owoce, bomby).
- ScoreManager - klasa, która odpowiada za zarządzanie wynikiem gracza i zapisywaniem go do pliku.
- Sword - klasa, która reprezentuje miecz gracza.

Naszym zadaniem będzie zaimplementowanie kilku metod w klasie GameManager oraz w Sliceable, które pozwolą na poprawne działanie gry.

Przejdźmy do pliku GameManager.py i zaimplementujmy metodę `handle_play`. Ta metoda będzie odpowiadała za główną część gdy, czyli obsługę zdarzeń, rysowanie obiektów na ekranie, itp.

Mamy już dedykowane miejsce pod tą metodę:

```python
def handle_play(self):
    pass
    # TODO
```

Słowo kluczowe pass oznacza, że metoda jest pusta i nie robi nic. Naszym zadaniem będzie zaimplementowanie tej metody, tak aby obsługiwała zdarzenia w grze.

Na początku wypełniamy cały ekran kolorem czarnym. W tym celu używamy metody fill, która wypełnia ekran podanym kolorem:

self.screen.fill((0, 0, 0))

Kilka słów o systemie RGB (Red, Green, Blue). Kolory w komputerze są reprezentowane za pomocą trzech składowych: czerwonej, zielonej i niebieskiej. Każda ze składowych ma wartość od 0 do 255, gdzie 0 oznacza brak koloru, a 255 pełny kolor. W naszym przypadku wypełniamy ekran kolorem czarnym, czyli brak czerwonego, zielonego i niebieskiego koloru.

Potem chcemy poinformować użytkownika o jego wyniku i czasie gry. Tworzymy dwie zmienne font i text, które przechowują informacje o wyniku i czasie gry. Następnie rysujemy te informacje na ekranie:

font = pygame.font.Font(None, 36)
text = font.render(f"Wynik: {self.score_manager.score}", True, (255, 255, 255))
self.screen.blit(text, (10, 10))

font_time = pygame.font.Font(None, 36)
text_time = font_time.render(f"Czas: {self.game_time:.2f}", True, (255, 255, 255))
self.screen.blit(text_time, (650, 10))

Następnie odliczamy czas gry i sprawdzamy, czy czas dobiegł do zera. Jeżeli tak, to zmieniamy stan gry na GAME_OVER:

self.game_time -= 1 / 60
if self.game_time <= 0:
    self.game_state = GameState.GAME_OVER

Ważną częścią gry jest obsługa zdarzeń. Do tego celu używamy pętli for, która przechodzi przez wszystkie zdarzenia, które wystąpiły od ostatniego wywołania metody handle_play. Sprawdzamy, czy użytkownik kliknął przycisk od zamykania okna, jeżeli tak, to ustawiamy flagę running na False:

for event in pygame.event.get():
    if event.type == pygame.QUIT:
        self.running = False

Kolejną rzeczą, którą sprawdzamy to to czy użytkownik ma przyciśnięty lewy przycisk myszy. Możemy to zrobić za pomocą metody get_pressed, która zwraca listę przycisków myszy. Pierwszy element tej listy to lewy przycisk myszy. Wartość 1 oznacza, że przycisk jest wciśnięty, a 0, że nie jest.
Jeżeli tak, to ustawiamy zmienną is_clicked na True:

mouse_buttons = pygame.mouse.get_pressed()
is_clicked = mouse_buttons[0]

Musimy to wiedzieć, jeżeli chcemy narysować smugę miecza na ekranie. Smuga ma się tylko pojawić, jeżeli gracz trzyma lewy przycisk myszy.

self.sword.update()
self.sword.draw(self.screen, is_clicked)

W naszej grze oczywiście nie może zabraknąć owoców. Generujemy nowe owoce co 1-3 sekundy, które będą wylatywały z rogów ekranu. Najpierw sprawdzamy, czy minęło wystarczająco dużo czasu od ostatniego wygenerowania owocu:

current_time = pygame.time.get_ticks()
if current_time - self.last_fruit_time > random.randint(1000, 3000):
    self.sliceables.append(Sliceable())
    self.last_fruit_time = current_time

Ten kod sprawdza czy minęła jedna do trzech sekund od ostatniego wygenerowania owocu. Jeżeli tak, to generujemy nowy owoc i dodajemy go do listy sliceables.

Następnie sprawdzamy kolizje z obiektami, które można przeciąć:

slashed_bomb = False
for sliceable in self.sliceables:
    slashed_bomb = sliceable.check_collision(self.score_manager, is_clicked)
    sliceable.update()
    sliceable.draw(self.screen)

Oprócz tego sprawdzamy, czy obiekty spadły poza planszę. Jeżeli tak, to usuwamy je z listy sliceables bo nie są już potrzebne:

    if sliceable.is_off_screen(self.screen.get_height()):
        self.sliceables.remove(sliceable)

Na koniec sprawdzamy, czy gracz przeciął bombę. Jeżeli tak, to zmieniamy stan gry na GAME_OVER:

if slashed_bomb:
    self.game_state = GameState.GAME_OVER

Cała metoda handle_play powinna wyglądać tak:

def handle_play(self):

    self.screen.fill((0, 0, 0))

    # Wypisanie wyniku i czasu
    font = pygame.font.Font(None, 36)
    text = font.render(f"Wynik: {self.score_manager.score}", True, (255, 255, 255))
    self.screen.blit(text, (10, 10))

    font_time = pygame.font.Font(None, 36)
    text_time = font_time.render(f"Czas: {self.game_time:.2f}", True, (255, 255, 255))
    self.screen.blit(text_time, (650, 10))

    # Odliczanie czasu
    self.game_time -= 1 / 60
    if self.game_time <= 0:
        self.game_state = GameState.GAME_OVER

    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            self.running = False

    mouse_buttons = pygame.mouse.get_pressed()
    is_clicked = mouse_buttons[0]

    # Rysowanie miecza
    self.sword.update()
    self.sword.draw(self.screen, is_clicked)

    # Generacja owoców od 1 do 3 sekund
    current_time = pygame.time.get_ticks()
    if current_time - self.last_fruit_time > random.randint(1000, 3000):
        self.sliceables.append(Sliceable())
        self.last_fruit_time = current_time

    slashed_bomb = False
    for sliceable in self.sliceables:
        slashed_bomb = sliceable.check_collision(self.score_manager, is_clicked)
        sliceable.update()
        sliceable.draw(self.screen)

        # Usunięcie obiektów które spadają poza planszę
        if sliceable.is_off_screen(self.screen.get_height()):
            self.sliceables.remove(sliceable)

    if slashed_bomb:
        self.game_state = GameState.GAME_OVER

To jest koniec impementacji tej metody, ale została jeszcze jedna rzecz do zrobienia. Musimy zaimplementować metodę check_collision w klasie Sliceable, która będzie sprawdzała kolizje z obiektami, które można przeciąć. Przechodzimy więc do pliku Sliceable.py i zaimplementujemy tą metodę:

Na ten moment wygląda ona tak:

def check_collision(self, score_manager: ScoreManager, is_clicked):
    pass
    # TODO

Możemy zauważyć, że w definicji metody mamy dwa argumenty: score_manager i is_clicked. Pierwszy argument to obiekt klasy ScoreManager, który przechowuje wynik gracza. Drugi argument to zmienna typu bool, która mówi nam, czy gracz trzyma lewy przycisk myszy. score_manager pozwoli nam na dodawanie punktów za przecięcie owocu. Czym jest dwukropek? Dwukropek oznacza, że argument jest typowany, czyli mówimy Pythonowi, że oczekujemy obiektu klasy ScoreManager. Pozwala to na lepsze zrozumienie kodu i uniknięcie błędów.

Pierwsze co zrobimy to zapiszemy sobie pozycję kursora, czyli naszego miecza:

sword = pygame.mouse.get_pos()

Następnie obliczamy odległość między myszką a środkiem owocu. Możemy to zrobić za pomocą wzoru na odległość między dwoma punktami.

distance = ((sword[0] - self.x) ** 2 + (sword[1] - self.y) ** 2) ** 0.5

Jeżeli odległość jest mniejsza od sumy promienia owocu i promienia smugi miecza, to znaczy, że gracz przeciął owoc. Wtedy dodajemy punkty za przecięcie owocu, generujemy cząsteczki i zwracamy False. Jeżeli przecięliśmy bombę, to zwracamy True.

if distance <= self.radius + 3 and not self.is_already_sliced and is_clicked:
    self.is_already_sliced = True
    if self.is_bomb:
        return True
    else:
        score_manager.add_points(100)
        return False

Na koniec dodajemy jeszcze metodę generate_particles, która generuje cząsteczki po przecięciu owocu:

def generate_particles(self):
    for i in range(random.randint(3, 6)):
        self.particles.append(Particle(self.x, self.y, self.color, self.speed_x, self.speed_y))

Tutaj generujemy od 3 do 6 cząsteczek, które mają losowe prędkości i kierunek.

Dodajmy wywołanie tej metody do metody check_collision:

if distance <= self.radius + 3 and not self.is_already_sliced and is_clicked:
    self.is_already_sliced = True
    if self.is_bomb:
        return True
    else:
        score_manager.add_points(100)
        self.generate_particles()
        return False


Gotowa metoda wyglą w taki sposób:

def check_collision(self, score_manager: ScoreManager, is_clicked):
    # Oblicz odległość między myszką a środkiem owocu
    sword = pygame.mouse.get_pos()
    distance = ((sword[0] - self.x) ** 2 + (sword[1] - self.y) ** 2) ** 0.5
    
    # Sprawdź czy myszka jest w zasięgu owocu
    if distance <= self.radius + 3 and not self.is_already_sliced and is_clicked:
        self.is_already_sliced = True
        if self.is_bomb:
            return True  # Bomba
        else:
            score_manager.add_points(100)
            self.generate_particles()
            return False  # Owoc
    
def generate_particles(self):
    for i in range(random.randint(3, 6)):
        self.particles.append(Particle(self.x, self.y, self.color, self.speed_x, self.speed_y))

Możemy teraz uruchomić naszą grę i sprawdzić, czy wszystko działa poprawnie. Jeżeli wszystko działa, to gratulacje! Udało się zaimplementować grę Fruit Warrior w języku Python!

## Podsumowanie

Podczas tej lekcji uczestnicy zapoznali się z podstawowymi pojęciami związanymi z programowaniem obiektowym w języku Python. Uczniowie mieli okazję pomóc w implementacji gry Fruit Warrior, w której gracz zdobywa punkty poprzez cięcie owoców. Dzięki tej lekcji uczestnicy poznali podstawy programowania w języku Python i zaimplementowali kilka metod, które pozwoliły na poprawne działanie gry.

Jeżeli zostało jeszcze trochę czasu, to zachęcamy uczestników do eksperymentowania z kodem i dodawania nowych funkcji do gry. Można spróbować dodać nowe obiekty, takie jak owoce różnego koloru (przypominamy im o systemie RGB i pytamy jaki kolor powinna mieć np. pomarańcza), albo dodatkowe efekty w przypadku gdy gracz przetnie bombę.